"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const chalk_1 = __importDefault(require("chalk"));
const read_config_1 = __importDefault(require("../utils/read-config"));
const review_project_1 = __importDefault(require("../utils/review-project"));
const dependents_to_string_1 = __importDefault(require("../utils/dependents-to-string"));
module.exports = {
    name: 'addon-guard',
    description: 'Review project to ensure that no addon dependency conflicts are present.',
    works: 'insideProject',
    availableOptions: Object.freeze([
        {
            name: 'all',
            type: Boolean,
            default: false,
            description: 'whether to include all addons, including those without run-time conflicts',
        }
    ]),
    run(options) {
        const config = read_config_1.default(this.project);
        const summary = review_project_1.default(this.project, {
            ignoreAddons: config.ignoreAddons || [],
            runtimeOnly: !options.all,
            conflictsOnly: !options.all,
            skipCacheKeyDependencyChecks: config.skipCacheKeyDependencyChecks
        });
        const addons = summary.addons;
        const addonCount = Object.keys(addons).length;
        let duplicateAddons = {};
        if (options.all) {
            if (addonCount > 0) {
                this.ui.writeLine(chalk_1.default.white(`ember-cli-addon-guard identified ${addonCount} addons:\n`));
                for (const name in addons) {
                    const addonInstances = addons[name];
                    this.ui.writeLine(chalk_1.default.underline(`${name}`));
                    this.ui.writeLine(dependents_to_string_1.default(name, addonInstances));
                    if (Object.keys(addonInstances).length > 1) {
                        duplicateAddons[name] = addonInstances;
                    }
                }
            }
            else {
                this.ui.writeLine(chalk_1.default.green(`ember-cli-addon-guard could not identify any addons.`));
            }
        }
        else if (addonCount > 0) {
            duplicateAddons = addons;
        }
        if (Object.keys(duplicateAddons).length > 0 || summary.errors.length > 0) {
            if (Object.keys(duplicateAddons).length > 0) {
                this.ui.writeLine(chalk_1.default.red(`ember-cli-addon-guard determined that your application is dependent on multiple versions of the following run-time ${addonCount > 1 ? 'addons' : 'addon'}:\n`));
                for (const name in duplicateAddons) {
                    const addonInstances = duplicateAddons[name];
                    this.ui.writeLine(chalk_1.default.red(chalk_1.default.underline(`${name}`)));
                    this.ui.writeLine(chalk_1.default.red(dependents_to_string_1.default(name, addonInstances)));
                }
            }
            if (summary.errors.length > 0) {
                this.ui.writeLine(chalk_1.default.red(`ember-cli-addon-guard has identified the following errors:\n`));
                for (const error of summary.errors) {
                    this.ui.writeLine(chalk_1.default.red(error) + '\n');
                }
            }
        }
        else {
            this.ui.writeLine(chalk_1.default.green(`ember-cli-addon-guard found no problems with your project.`));
        }
    }
};
//# sourceMappingURL=addon-guard.js.map